# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendor invoice parser for United Natural Foods (UNFI)
"""

from __future__ import unicode_literals

import datetime

from rattail.vendors.invoices import InvoiceParser
from rattail.db.batch.vendorinvoice import VendorInvoiceRow
from rattail.csvutil import UnicodeDictReader, UnicodeReader
from rattail.gpc import GPC


class UnfiInvoiceParser(InvoiceParser):
    """
    Parser for UNFI CSV invoice files.
    """
    key = 'rattail.contrib.unfi'
    display = "United Natural Foods (UNFI)"
    vendor_key = 'unfi'

    def parse_invoice_date(self, data_path):
        with open(data_path, 'rb') as f:
            reader = UnicodeDictReader(f, encoding='latin_1')
            data = reader.next()
            return datetime.datetime.strptime(data['InvoiceDate'], '%m/%d/%Y').date()

    def parse_rows(self, data_path):
        with open(data_path, 'rb') as csv_file:

            # TODO: The following logic is largely copied from the Scan Genius
            # order parser (in `rattail_livnat.scangenius`).  I wonder if we
            # can abstract it into a generic yet custom CSV parser...?

            # We want to ignore the header section here.  However the use of
            # UnicodeDictReader below requires iteration, and yet also expects
            # the "beginning" of the file to contain the fieldnames.  To force
            # this all to play nicely, we calculate our offset and then seek to
            # it explicitly.  It's probably likely that there is a better way,
            # but this works.
            offset = self.find_details_offset(csv_file)
            csv_file.seek(offset)

            reader = UnicodeDictReader(csv_file, encoding='latin_1')
            for data in reader:

                # Only consider 'Detail' rows; this check is mostly for the
                # sake of ignoring 'Footer' rows.
                if data['RecType'] != 'Detail':
                    continue

                row = VendorInvoiceRow()
                row.upc = GPC(data['UPC'].replace('-', ''))
                row.vendor_code = data['ProductID']
                row.brand_name = data['Brand']
                row.description = data['Description']
                row.size = data['Size']
                row.case_quantity = self.int_(data['Servings'])
                row.ordered_cases = int(self.decimal(data['QuantityOrdered']))
                row.shipped_cases = self.int_(data['QuantityShipped'])
                row.unit_cost = self.decimal(data['NetPricePerUnit'])
                row.total_cost = self.decimal(data['ExtendedPrice'])
                row.case_cost = row.unit_cost * row.case_quantity

                yield row

    def find_details_offset(self, csv_file):
        """
        Find the character offset of the details data within a CSV file.
        """
        # TODO: The following logic is largely copied from the Scan Genius
        # order parser (in ``rattail_livnat.scangenius``).  I wonder if we can
        # abstract it into a generic yet custom CSV parser...?

        # Calculate the offset of the details section, based on the sum
        # character length of the header section (which is 2 lines long).
        reader = UnicodeReader(csv_file, encoding='latin_1')
        offset = 0
        for line, row in enumerate(csv_file, 1):
            offset += len(row)
            if line == 2:
                break
        return offset



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import argparse
from unittest import TestCase

from mock import Mock, patch

from rattail.core import Object
from rattail.commands import importing
from rattail.importing import Importer, ImportHandler
from rattail.importing.rattail import FromRattailToRattail
from rattail.config import RattailConfig
from rattail.tests.importing import ImporterTester
from rattail.tests.importing.test_handlers import MockImportHandler
from rattail.tests.importing.test_importers import MockImporter
from rattail.tests.importing.test_rattail import DualRattailTestCase


class MockImport(importing.ImportSubcommand):

    def get_handler_factory(self):
        return MockImportHandler


class TestImportSubcommand(TestCase):

    # TODO: lame, here only for coverage
    def test_parent_name(self):
        parent = Object(name='milo')
        command = importing.ImportSubcommand(parent=parent)
        self.assertEqual(command.parent_name, 'milo')

    def test_get_handler_factory(self):
        command = importing.ImportSubcommand()
        self.assertRaises(NotImplementedError, command.get_handler_factory)
        command.handler_spec = 'rattail.importing.rattail:FromRattailToRattail'
        factory = command.get_handler_factory()
        self.assertIs(factory, FromRattailToRattail)

    def test_handler_spec_attr(self):
        # default is None
        command = importing.ImportSubcommand()
        self.assertIsNone(command.handler_spec)

        # can't get a handler without a spec
        self.assertRaises(NotImplementedError, command.get_handler)

        # but may be specified with init kwarg
        command = importing.ImportSubcommand(handler_spec='rattail.importing:ImportHandler')
        self.assertEqual(command.handler_spec, 'rattail.importing:ImportHandler')

        # now we can get a handler
        handler = command.get_handler()
        self.assertIsInstance(handler, ImportHandler)

    def test_get_handler(self):

        # no config
        command = MockImport()
        handler = command.get_handler()
        self.assertIs(type(handler), MockImportHandler)
        self.assertIsNone(handler.config)

        # with config
        config = RattailConfig()
        command = MockImport(config=config)
        handler = command.get_handler()
        self.assertIs(type(handler), MockImportHandler)
        self.assertIs(handler.config, config)

        # dry run
        command = MockImport()
        handler = command.get_handler()
        self.assertFalse(handler.dry_run)
        handler = command.get_handler(dry_run=True)
        self.assertTrue(handler.dry_run)
        args = argparse.Namespace(dry_run=True)
        handler = command.get_handler(args=args)
        self.assertTrue(handler.dry_run)

    def test_add_parser_args(self):

        # adding the args throws no error..(basic coverage)
        command = importing.ImportSubcommand()
        parser = argparse.ArgumentParser()
        command.add_parser_args(parser)

        # confirm default values
        args = parser.parse_args([])
        self.assertIsNone(args.start_date)
        self.assertIsNone(args.end_date)
        self.assertTrue(args.create)
        self.assertIsNone(args.max_create)
        self.assertTrue(args.update)
        self.assertIsNone(args.max_update)
        self.assertFalse(args.delete)
        self.assertIsNone(args.max_delete)
        self.assertIsNone(args.max_total)
        self.assertEqual(args.batch_size, 200)
        self.assertFalse(args.warnings)
        self.assertFalse(args.dry_run)

    def test_batch_size_kwarg(self):
        command = importing.ImportSubcommand(handler_spec='rattail.importing:ImportHandler')
        parser = argparse.ArgumentParser()
        command.add_parser_args(parser)
        with patch.object(ImportHandler, 'get_importers', Mock(return_value={'Foo': Importer})):

            # importer default is 200
            args = parser.parse_args([])
            handler = command.get_handler(args=args)
            importer = handler.get_importer('Foo')
            self.assertEqual(importer.batch_size, 200)

            # but may be overridden with command line arg
            args = parser.parse_args(['--batch', '42'])
            handler = command.get_handler(args=args)
            importer = handler.get_importer('Foo')
            self.assertEqual(importer.batch_size, 42)


class TestImportSubcommandRun(TestCase, ImporterTester):

    sample_data = {
        '16oz': {'upc': '00074305001161', 'description': "Apple Cider Vinegar 16oz"},
        '32oz': {'upc': '00074305001321', 'description': "Apple Cider Vinegar 32oz"},
        '1gal': {'upc': '00074305011283', 'description': "Apple Cider Vinegar 1gal"},
    }

    def setUp(self):
        self.command = MockImport()
        self.handler = MockImportHandler()
        self.importer = MockImporter()

    def import_data(self, host_data=None, local_data=None, **kwargs):
        if host_data is None:
            host_data = self.sample_data
        if local_data is None:
            local_data = self.sample_data

        models = kwargs.pop('models', [])
        kwargs.setdefault('dry_run', False)

        kw = {
            'warnings': False,
            'create': None,
            'max_create': None,
            'update': None,
            'max_update': None,
            'delete': None,
            'max_delete': None,
            'max_total': None,
            'batchcount': None,
            'progress': None,
        }
        kw.update(kwargs)
        args = argparse.Namespace(models=models, **kw)

        # must modify our importer in-place since we need the handler to return
        # that specific instance, below (because the host/local data context
        # managers reference that instance directly)
        self.importer._setup(**kwargs)
        with patch.object(self.command, 'get_handler', Mock(return_value=self.handler)):
            with patch.object(self.handler, 'get_importer', Mock(return_value=self.importer)):
                with self.host_data(host_data):
                    with self.local_data(local_data):
                        self.command.run(args)

        if self.handler._result:
            self.result = self.handler._result['Product']
        else:
            self.result = [], [], []

    def test_create(self):
        local = self.copy_data()
        del local['32oz']
        self.import_data(local_data=local)
        self.assert_import_created('32oz')
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_update(self):
        local = self.copy_data()
        local['16oz']['description'] = "wrong description"
        self.import_data(local_data=local)
        self.assert_import_created()
        self.assert_import_updated('16oz')
        self.assert_import_deleted()

    def test_delete(self):
        local = self.copy_data()
        local['bogus'] = {'upc': '00000000000000', 'description': "Delete Me"}
        self.import_data(local_data=local, delete=True)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted('bogus')

    def test_duplicate(self):
        host = self.copy_data()
        host['32oz-dupe'] = host['32oz']
        self.import_data(host_data=host)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_max_create(self):
        local = self.copy_data()
        del local['16oz']
        del local['1gal']
        self.import_data(local_data=local, max_create=1)
        self.assert_import_created('16oz')
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_max_total_create(self):
        local = self.copy_data()
        del local['16oz']
        del local['1gal']
        self.import_data(local_data=local, max_total=1)
        self.assert_import_created('16oz')
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_max_update(self):
        local = self.copy_data()
        local['16oz']['description'] = "wrong"
        local['1gal']['description'] = "wrong"
        self.import_data(local_data=local, max_update=1)
        self.assert_import_created()
        self.assert_import_updated('16oz')
        self.assert_import_deleted()

    def test_max_total_update(self):
        local = self.copy_data()
        local['16oz']['description'] = "wrong"
        local['1gal']['description'] = "wrong"
        self.import_data(local_data=local, max_total=1)
        self.assert_import_created()
        self.assert_import_updated('16oz')
        self.assert_import_deleted()

    def test_max_delete(self):
        local = self.copy_data()
        local['bogus1'] = {'upc': '00000000000001', 'description': "Delete Me"}
        local['bogus2'] = {'upc': '00000000000002', 'description': "Delete Me"}
        self.import_data(local_data=local, delete=True, max_delete=1)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted('bogus1')

    def test_max_total_delete(self):
        local = self.copy_data()
        local['bogus1'] = {'upc': '00000000000001', 'description': "Delete Me"}
        local['bogus2'] = {'upc': '00000000000002', 'description': "Delete Me"}
        self.import_data(local_data=local, delete=True, max_total=1)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted('bogus1')

    def test_dry_run(self):
        local = self.copy_data()
        del local['32oz']
        local['16oz']['description'] = "wrong description"
        local['bogus'] = {'upc': '00000000000000', 'description': "Delete Me"}
        self.import_data(local_data=local, delete=True, dry_run=True)
        # TODO: maybe need a way to confirm no changes actually made due to dry
        # run; currently results still reflect "proposed" changes.  this rather
        # bogus test is here just for coverage sake
        self.assert_import_created('32oz')
        self.assert_import_updated('16oz')
        self.assert_import_deleted('bogus')


class TestImportRattail(DualRattailTestCase):

    def make_command(self, **kwargs):
        kwargs.setdefault('config', self.config)
        return importing.ImportRattail(**kwargs)

    def test_get_handler_factory(self):

        # default handler
        config = RattailConfig()
        command = self.make_command(config=config)
        Handler = command.get_handler_factory()
        self.assertIs(Handler, FromRattailToRattail)

    def test_get_handler_kwargs(self):
        command = self.make_command()
        kwargs = command.get_handler_kwargs()
        self.assertEqual(kwargs, {})

        args = argparse.Namespace(dbkey='other')
        kwargs = command.get_handler_kwargs(args=args)
        self.assertEqual(len(kwargs), 2)
        self.assertIs(kwargs['args'], args)
        self.assertEqual(kwargs['dbkey'], 'other')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Basic Models for Batches

Actually the classes in this module are not true models but rather are mixins,
which provide the common columns etc. for batch tables.
"""

from __future__ import unicode_literals

import os
import datetime

import sqlalchemy as sa
from sqlalchemy.orm import relationship, object_session
from sqlalchemy.ext.declarative import declared_attr

from rattail.db.core import uuid_column
from rattail.db.types import GPCType
from rattail.db.model import User, Product


class BatchMixin(object):
    """
    Mixin for all (new-style) batch classes.

    .. note::
       This is all still very experimental.
    """

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__()

    @classmethod
    def __default_table_args__(cls):
        return (
            sa.ForeignKeyConstraint(['created_by_uuid'], ['user.uuid'],
                                    name='{0}_fk_created_by'.format(cls.__tablename__)),
            sa.ForeignKeyConstraint(['cognized_by_uuid'], ['user.uuid'],
                                    name='{0}_fk_cognized_by'.format(cls.__tablename__)), 
            sa.ForeignKeyConstraint(['executed_by_uuid'], ['user.uuid'],
                                    name='{0}_fk_executed_by'.format(cls.__tablename__)),
            )

    @declared_attr
    def batch_key(cls):
        return cls.__tablename__

    uuid = uuid_column()

    created = sa.Column(sa.DateTime(), nullable=False, default=datetime.datetime.utcnow, doc="""
Date and time when the batch was first created.
""")

    created_by_uuid = sa.Column(sa.String(length=32), nullable=False)

    @declared_attr
    def created_by(cls):
        return relationship(User,
                            primaryjoin=lambda: User.uuid == cls.created_by_uuid,
                            foreign_keys=lambda: [cls.created_by_uuid], doc="""
Reference to the :class:`User` who first created the batch.
""")

    cognized = sa.Column(sa.DateTime(), nullable=True, doc="""
Date and time when the batch data was last cognized.
""")

    cognized_by_uuid = sa.Column(sa.String(length=32), nullable=True)

    @declared_attr
    def cognized_by(cls):
        return relationship(User,
                            primaryjoin=lambda: User.uuid == cls.cognized_by_uuid,
                            foreign_keys=lambda: [cls.cognized_by_uuid], doc="""
Reference to the :class:`User` who last cognized the batch data.
""")

    rowcount = sa.Column(sa.Integer(), nullable=True, doc="""
Cached row count for the batch.  No guarantees perhaps, but should be accurate.
""")

    executed = sa.Column(sa.DateTime(), nullable=True, doc="""
Date and time when the batch was (last) executed.
""")

    executed_by_uuid = sa.Column(sa.String(length=32), nullable=True)

    @declared_attr
    def executed_by(cls):
        return relationship(User,
                            primaryjoin=lambda: User.uuid == cls.executed_by_uuid,
                            foreign_keys=lambda: [cls.executed_by_uuid], doc="""
Reference to the :class:`User` who (last) executed the batch.
""")

    purge = sa.Column(sa.Date(), nullable=True, doc="""
Date after which the batch may be purged.
""")

    def __repr__(self):
        return "{}(uuid={})".format(
            self.__class__.__name__,
            repr(self.uuid))


class FileBatchMixin(BatchMixin):
    """
    Mixin for all (new-style) batch classes which involve a file upload as
    their first step.

    .. note::
       This is all still very experimental.
    """

    filename = sa.Column(sa.String(length=255), nullable=False, doc="""
Base name of the file which was used as the data source.
""")

    def relative_filepath(self, config):
        """
        Returns the path for the data file, relative to the root folder for
        batch file storage.  This includes the batch key as the first segment.
        """
        if not self.uuid:
            object_session(self).flush()
        return os.path.join(self.batch_key, self.uuid[:2], self.uuid[2:])

    def filedir(self, config):
        """
        Returns the absolute path to the folder in which the data file resides.
        The config object determines the root path for such files, e.g.:

        .. code-block:: ini

           [rattail]
           batch.files = /path/to/batch/files

        Within this root path, a more complete path is generated using the
        :attr:`BatchMixin.key` and the :attr:`BatchMixin.uuid` values.
        """
        batchdir = config.batch_filedir()
        return os.path.abspath(os.path.join(batchdir, self.relative_filepath(config)))

    def absolute_filepath(self, config):
        """
        Return the absolute path where the data file resides.  This is the path
        returned by :meth:`filedir()` with the batch filename joined to it.
        """
        return os.path.join(self.filedir(config), self.filename)

    # for convenience
    filepath = absolute_filepath

    def filesize(self, config):
        """
        Returns the size of the data file in bytes.
        """
        path = self.filepath(config)
        return os.path.getsize(path)

    def write_file(self, config, contents):
        """
        Save a data file for the batch to the location specified by
        :meth:`filepath()`.
        """
        filedir = self.filedir(config)
        if not os.path.exists(filedir):
            os.makedirs(filedir)
        with open(os.path.join(filedir, self.filename), 'wb') as f:
            f.write(contents)

    def delete_data(self, config):
        """
        Delete the data file and folder for the batch.
        """
        path = self.filepath(config)
        if os.path.exists(path):
            os.remove(path)
        path = self.filedir(config)
        if os.path.exists(path):
            os.rmdir(path)


class BatchRowMixin(object):
    """
    Mixin for all (new-style) batch row classes.

    .. note::
       This is all still very experimental.
    """

    uuid = uuid_column()

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__()

    @classmethod
    def __default_table_args__(cls):
        batch_table = cls.__batch_class__.__tablename__
        row_table = cls.__tablename__
        return (
            sa.ForeignKeyConstraint(['batch_uuid'], ['{0}.uuid'.format(batch_table)],
                                    name='{0}_fk_batch_uuid'.format(row_table)),
            )

    STATUS = {}

    batch_uuid = sa.Column(sa.String(length=32), nullable=False)

    @declared_attr
    def batch(cls):
        batch_class = cls.__batch_class__
        row_class = cls

        # Must establish `Batch.data_rows` here instead of from within `Batch`
        # itself, because the row class doesn't yet exist when that happens.
        batch_class.data_rows = relationship(row_class, back_populates='batch',
                                             order_by=lambda: row_class.sequence,
                                             cascade='all, delete-orphan', doc="""
Collection of data rows for the batch.

.. note::
   I would prefer for this attribute to simply be named "rows" instead of
   "data_rows", but unfortunately (as of this writing) "rows" is essentially a
   reserved word in FormAlchemy.
""")

        # Now, here's the `BatchRow.batch` reference.
        return relationship(batch_class, back_populates='data_rows', doc="""
Reference to the parent batch to which the row belongs.
""")

    sequence = sa.Column(sa.Integer(), nullable=False, doc="""
Sequence number of the row within the batch.  This number should be from 1 to
the actual number of rows in the batch.
""")

    status_code = sa.Column(sa.Integer(), nullable=True, doc="""
Status code for the data row.  This indicates whether the row's product could
be found in the system, etc.  Ultimately the meaning of this is defined by each
particular batch type.
""")

    status_text = sa.Column(sa.String(length=255), nullable=True, doc="""
Short description of row status.  Ultimately the meaning and use of this is
defined by each particular batch type.
""")

    removed = sa.Column(sa.Boolean(), nullable=False, default=False, doc="""
Flag to indicate a row has been removed from the batch.
""")


class ProductBatchRowMixin(BatchRowMixin):
    """
    Mixin for all row classes of (new-style) batches which pertain to products.

    .. note::
       This is all still very experimental.
    """

    @classmethod
    def __default_table_args__(cls):
        batch_table = cls.__batch_class__.__tablename__
        row_table = cls.__tablename__
        return (
            sa.ForeignKeyConstraint(['batch_uuid'], ['{0}.uuid'.format(batch_table)],
                                    name='{0}_fk_batch'.format(row_table)),
            sa.ForeignKeyConstraint(['product_uuid'], ['product.uuid'],
                                    name='{0}_fk_product'.format(row_table)),
            )

    upc = sa.Column(GPCType(), nullable=True, doc="""
UPC of the product whose authz cost should be changed.
""")

    product_uuid = sa.Column(sa.String(length=32), nullable=True)

    @declared_attr
    def product(self):
        return relationship(Product, doc="""
Reference to the :class:`Product` with which the row is associated, if any.
""")

    brand_name = sa.Column(sa.String(length=100), nullable=True, doc="""
Brand name of the product.
""")

    description = sa.Column(sa.String(length=255), nullable=True, doc="""
Description of the product.
""")

    size = sa.Column(sa.String(length=255), nullable=True, doc="""
Size of the product, as string.
""")



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

from unittest import TestCase

from mock import patch

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm.exc import MultipleResultsFound

from rattail.importing import sqlalchemy as saimport


class Widget(object):
    pass

metadata = sa.MetaData()

widget_table = sa.Table('widgets', metadata,
                        sa.Column('id', sa.Integer(), primary_key=True),
                        sa.Column('description', sa.String(length=50)))

widget_mapper = orm.mapper(Widget, widget_table)

WIDGETS = [
    {'id': 1, 'description': "Main Widget"},
    {'id': 2, 'description': "Other Widget"},
    {'id': 3, 'description': "Other Widget"},
]


class TestFromSQLAlchemy(TestCase):

    def test_query(self):
        Session = orm.sessionmaker(bind=sa.create_engine('sqlite://'))
        session = Session()
        importer = saimport.FromSQLAlchemy(host_session=session,
                                           host_model_class=Widget)
        self.assertEqual(unicode(importer.query()),
                         "SELECT widgets.id AS widgets_id, widgets.description AS widgets_description \n"
                         "FROM widgets")


class TestToSQLAlchemy(TestCase):

    def setUp(self):
        engine = sa.create_engine('sqlite://')
        metadata.create_all(bind=engine)
        Session = orm.sessionmaker(bind=engine)
        self.session = Session()
        for data in WIDGETS:
            widget = Widget()
            for key, value in data.iteritems():
                setattr(widget, key, value)
            self.session.add(widget)
        self.session.commit()

    def tearDown(self):
        self.session.close()
        del self.session

    def make_importer(self, **kwargs):
        kwargs.setdefault('model_class', Widget)
        return saimport.ToSQLAlchemy(**kwargs)

    def test_model_mapper(self):
        importer = self.make_importer()
        self.assertIs(importer.model_mapper, widget_mapper)

    def test_model_table(self):
        importer = self.make_importer()
        self.assertIs(importer.model_table, widget_table)

    def test_simple_fields(self):
        importer = self.make_importer()
        self.assertEqual(importer.simple_fields, ['id', 'description'])

    def test_get_single_local_object(self):

        # simple key
        importer = self.make_importer(key='id', session=self.session)
        widget = importer.get_single_local_object((1,))
        self.assertEqual(widget.id, 1)
        self.assertEqual(widget.description, "Main Widget")

        # compound key
        importer = self.make_importer(key=('id', 'description'), session=self.session)
        widget = importer.get_single_local_object((1, "Main Widget"))
        self.assertEqual(widget.id, 1)
        self.assertEqual(widget.description, "Main Widget")

        # widget not found
        importer = self.make_importer(key='id', session=self.session)
        self.assertIsNone(importer.get_single_local_object((42,)))

        # multiple widgets found
        importer = self.make_importer(key='description', session=self.session)
        self.assertRaises(MultipleResultsFound, importer.get_single_local_object, ("Other Widget",))

    def test_create_object(self):
        importer = self.make_importer(key='id', session=self.session)
        widget = importer.create_object((42,), {'id': 42, 'description': "Latest Widget"})
        self.assertIn(widget, self.session)
        self.assertIn(widget, self.session.new)
        self.assertEqual(widget.id, 42)
        self.assertEqual(widget.description, "Latest Widget")

    def test_delete_object(self):
        widget = self.session.query(Widget).get(1)
        self.assertIn(widget, self.session)
        importer = self.make_importer(session=self.session)
        self.assertTrue(importer.delete_object(widget))
        self.assertIn(widget, self.session)
        self.assertIn(widget, self.session.deleted)
        self.session.flush()
        self.assertIsNone(self.session.query(Widget).get(1))

    def test_cache_model(self):
        importer = self.make_importer(key='id', session=self.session)
        cache = importer.cache_model(Widget, key='id')
        self.assertEqual(len(cache), 3)
        for i in range(1, 4):
            self.assertIn(i, cache)
            self.assertIsInstance(cache[i], Widget)
            self.assertEqual(cache[i].id, i)
            self.assertEqual(cache[i].description, WIDGETS[i-1]['description'])

    def test_cache_local_data(self):
        importer = self.make_importer(key='id', session=self.session)
        cache = importer.cache_local_data()
        self.assertEqual(len(cache), 3)
        for i in range(1, 4):
            self.assertIn((i,), cache)
            cached = cache[(i,)]
            self.assertIsInstance(cached, dict)
            self.assertIsInstance(cached['object'], Widget)
            self.assertEqual(cached['object'].id, i)
            self.assertEqual(cached['object'].description, WIDGETS[i-1]['description'])
            self.assertIsInstance(cached['data'], dict)
            self.assertEqual(cached['data']['id'], i)
            self.assertEqual(cached['data']['description'], WIDGETS[i-1]['description'])

    # TODO: lame
    def test_flush_session(self):
        importer = self.make_importer(fields=['id'], key='id', session=self.session, flush_session=True)
        widget = Widget()
        widget.id = 1
        widget, original = importer.update_object(widget, {'id': 1}), widget
        self.assertIs(widget, original)

    def test_flush_create_update(self):
        importer = saimport.ToSQLAlchemy(fields=['id'], session=self.session)
        with patch.object(self.session, 'flush') as flush:
            importer.flush_create_update()
            flush.assert_called_once_with()



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

from unittest import TestCase

import sqlalchemy as sa
from fixture import TempIO

from rattail import db
from rattail.config import RattailConfig


class TestSession(TestCase):

    def test_init_rattail_config(self):
        session = db.Session()
        self.assertIsNone(session.rattail_config)
        session.close()

        config = object()
        session = db.Session(rattail_config=config)
        self.assertIs(session.rattail_config, config)
        session.close()

    def test_init_record_changes(self):
        self.assertFalse(db.Session.kw['rattail_record_changes'])

        session = db.Session()
        self.assertFalse(session.rattail_record_changes)
        session.close()

        session = db.Session(rattail_record_changes=True)
        self.assertTrue(session.rattail_record_changes)
        session.close()

        engine = sa.create_engine('sqlite://')
        engine.rattail_record_changes = True
        session = db.Session(bind=engine)
        self.assertTrue(session.rattail_record_changes)
        session.close()


class TestConfigExtension(TestCase):

    def setUp(self):
        self.tempio = TempIO()

    def tearDown(self):
        db.Session.configure(bind=None, rattail_config=None)
        self.tempio = None

    def test_configure_empty(self):
        config = RattailConfig()
        self.assertRaises(AttributeError, getattr, config, 'rattail_engines')
        self.assertRaises(AttributeError, getattr, config, 'rattail_engine')
        self.assertIsNone(db.Session.kw['bind'])
        self.assertIsNone(db.Session.kw['rattail_config'])

        db.ConfigExtension().configure(config)
        self.assertEqual(config.rattail_engines, {})
        self.assertIsNone(config.rattail_engine)
        self.assertIs(db.Session.kw['rattail_config'], config)

    def test_configure_connections(self):
        default_path = self.tempio.putfile('default.sqlite', '')
        default_url = 'sqlite:///{}'.format(default_path)
        host_path = self.tempio.putfile('host.sqlite', '')
        host_url = 'sqlite:///{}'.format(host_path)

        config = RattailConfig()
        config.set('rattail.db', 'keys', 'default, host')
        config.set('rattail.db', 'default.url', default_url)
        config.set('rattail.db', 'host.url', host_url)
        db.ConfigExtension().configure(config)
        self.assertEqual(len(config.rattail_engines), 2)
        self.assertEqual(unicode(config.rattail_engines['default'].url), default_url)
        self.assertEqual(unicode(config.rattail_engines['host'].url), host_url)
        self.assertEqual(unicode(config.rattail_engine.url), default_url)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

from unittest import TestCase

import sqlalchemy as sa
from mock import patch
from fixture import TempIO

from rattail.db import model, Session, SessionBase, auth
from rattail.importing import rattail as rattail_importing
from rattail.tests import RattailMixin, RattailTestCase
from rattail.tests.importing import ImporterTester


class DualRattailMixin(RattailMixin):

    def setup_rattail(self):
        super(DualRattailMixin, self).setup_rattail()

        if 'host' not in self.config.rattail_engines:
            self.config.rattail_engines['host'] = sa.create_engine('sqlite://')

        self.host_engine = self.config.rattail_engines['host']
        self.config.setdefault('rattail.db', 'keys', 'default, host')
        self.config.setdefault('rattail.db', 'host.url', unicode(self.host_engine.url))
        model = self.get_rattail_model()
        model.Base.metadata.create_all(bind=self.host_engine)
        self.host_session = Session(bind=self.host_engine)
        
    def teardown_rattail(self):
        super(DualRattailMixin, self).teardown_rattail()

        self.host_session.close()
        model = self.get_rattail_model()
        model.Base.metadata.drop_all(bind=self.config.rattail_engines['host'])

        if hasattr(self, 'tempio'):
            self.tempio = None


class DualRattailTestCase(DualRattailMixin, TestCase):
    pass


class TestFromRattailHandler(RattailTestCase, ImporterTester):
    handler_class = rattail_importing.FromRattailHandler
        
    def test_make_host_session(self):
        handler = self.make_handler()
        session = handler.make_host_session()
        self.assertIsInstance(session, SessionBase)
        self.assertIs(session.bind, self.config.rattail_engine)


class TestFromRattailToRattail(DualRattailTestCase, ImporterTester):
    handler_class = rattail_importing.FromRattailToRattail

    def test_host_title(self):
        handler = self.make_handler()
        self.assertEqual(handler.host_title, "Rattail (host)")

    # TODO
    def test_default_keys(self):
        handler = self.make_handler()
        handler.get_default_keys()

    def test_make_session(self):
        handler = self.make_handler()
        session = handler.make_session()
        self.assertIsInstance(session, SessionBase)
        self.assertIs(session.bind, self.config.rattail_engine)

    def test_make_host_session(self):

        # default is 'host'
        handler = self.make_handler()
        session = handler.make_host_session()
        self.assertIsInstance(session, SessionBase)
        self.assertIs(session.bind, self.host_engine)

        # invalid dbkey
        handler = self.make_handler(dbkey='other')
        self.assertRaises(KeyError, handler.make_host_session)

        # alternate dbkey
        self.config.rattail_engines['other'] = self.config.rattail_engines['host']
        handler = self.make_handler(dbkey='other')
        session = handler.make_host_session()
        self.assertIsInstance(session, SessionBase)
        self.assertIs(session.bind, self.host_engine)


class TestFromRattail(DualRattailTestCase):

    def make_importer(self, model_class=None, **kwargs):
        kwargs.setdefault('host_session', self.host_session)
        importer = rattail_importing.FromRattail(self.config, **kwargs)
        if model_class:
            importer.model_class = model_class
        return importer

    def test_host_model_class(self):
        importer = self.make_importer()
        self.assertIsNone(importer.model_class)
        self.assertIsNone(importer.host_model_class)
        importer = self.make_importer(model.Product)
        self.assertIs(importer.host_model_class, model.Product)

    def test_query(self):
        importer = self.make_importer(model.Product)
        importer.query()

    def test_normalize_host_object(self):
        importer = self.make_importer(model.Product)
        product = model.Product()
        with patch.object(importer, 'normalize_local_object') as normalize_local:
            normalize_local.return_value = {}
            data = importer.normalize_host_object(product)
            self.assertEqual(data, {})
            normalize_local.assert_called_once_with(product)
        self.assertEqual(data, importer.normalize_local_object(product))


class TestAdminUser(DualRattailTestCase):

    importer_class = rattail_importing.AdminUserImporter

    def make_importer(self, **kwargs):
        kwargs.setdefault('session', self.session)
        return self.importer_class(**kwargs)

    def get_admin(self):
        return auth.administrator_role(self.session)

    def test_normalize_host_object(self):
        importer = self.make_importer()
        user = model.User()
        data = importer.normalize_host_object(user)
        self.assertFalse(data['admin'])
        user.roles.append(self.get_admin())
        data = importer.normalize_host_object(user)
        self.assertTrue(data['admin'])



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import datetime

import sqlalchemy as sa

from rattail.db import model
from rattail.db.batches import util
from rattail.tests import DataTestCase


class TestPurgeBatches(DataTestCase):

    def setUp(self):
        super(TestPurgeBatches, self).setUp()

        batch = model.Batch(purge=datetime.date(2014, 1, 1))
        batch.add_column('F01')
        batch.add_column('F02')
        self.session.add(batch)
        batch.create_table()

        batch = model.Batch(purge=datetime.date(2014, 2, 1))
        batch.add_column('F01')
        batch.add_column('F02')
        self.session.add(batch)
        batch.create_table()

        batch = model.Batch(purge=datetime.date(2014, 3, 1))
        batch.add_column('F01')
        batch.add_column('F02')
        self.session.add(batch)
        batch.create_table()

        self.session.commit()

    def get_batch_tables_metadata(self):
        def batch_tables(name, metadata):
            return util.batch_pattern.match(name)
        metadata = sa.MetaData(bind=self.engine)
        metadata.reflect(only=batch_tables)
        return metadata

    def test_purging_honors_batch_purge_dates(self):
        self.assertEqual(self.session.query(model.Batch).count(), 3)
        self.assertEqual(util.purge_batches(effective_date=datetime.date(2014, 1, 15)), 1)
        self.assertEqual(self.session.query(model.Batch).count(), 2)
        self.assertEqual(self.session.query(sa.func.min(model.Batch.purge)).scalar(), datetime.date(2014, 2, 1))

    def test_purging_everything_does_just_that(self):
        self.assertEqual(self.session.query(model.Batch).count(), 3)
        self.assertEqual(util.purge_batches(purge_everything=True), 3)
        self.assertEqual(self.session.query(model.Batch).count(), 0)

    # TODO: The next two tests each work if only one is enabled...but if both
    # are enabled, one will fail.  This needs more investigation, but one
    # possible cause is the "corruption" of Base.metadata when Batch.rowclass
    # is accessed?  In particular it seems *not* to be a SQLite problem, as it
    # occurred when using a PostgreSQL engine as well.

    # def test_purging_does_not_leave_orphaned_tables(self):
    #     self.assertEqual(self.session.query(model.Batch).count(), 3)
    #     self.assertEqual(util.purge_batches(purge_everything=True), 3)
    #     self.assertEqual(self.session.query(model.Batch).count(), 0)
    #     metadata = self.get_batch_tables_metadata()
    #     self.assertEqual(len(metadata.tables), 0)

    # def test_purging_does_not_delete_previously_orphaned_tables(self):
    #     metadata = self.get_batch_tables_metadata()
    #     self.assertEqual(len(metadata.tables), 3)
    #     batch = self.session.query(model.Batch).first()
    #     batch.drop_table()
    #     self.assertEqual(self.session.query(model.Batch).count(), 3)
    #     metadata = self.get_batch_tables_metadata()
    #     self.assertEqual(len(metadata.tables), 2)



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Common email config objects
"""

from __future__ import unicode_literals, absolute_import

import sys
from traceback import format_exception

from rattail.mail import Email


class datasync_error_watcher_get_changes(Email):
    """
    When any datasync watcher thread encounters an error trying to get changes,
    this email is sent out.
    """
    default_subject = "Watcher failed to get changes"

    def sample_data(self, request):
        from rattail.datasync import DataSyncWatcher
        try:
            raise RuntimeError("Fake error for preview")
        except:
            exc_type, exc, traceback = sys.exc_info()
        watcher = DataSyncWatcher(self.config, 'test')
        watcher.consumes_self = True
        return {
            'watcher': watcher,
            'error': exc,
            'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
        }


class filemon_action_error(Email):
    """
    When any filemon thread encounters an error (and the retry attempts have
    been exhausted) then it will send out this email.
    """
    default_subject = "Error invoking action(s)"

    def sample_data(self, request):
        from rattail.filemon import Action
        action = Action(self.config)
        action.spec = 'rattail.filemon.actions:Action'
        action.retry_delay = 10
        try:
            raise RuntimeError("Fake error for preview")
        except:
            exc_type, exc, traceback = sys.exc_info()
        return {
            'path': '/tmp/foo.csv',
            'action': action,
            'attempts': 3,
            'error': exc,
            'traceback': ''.join(format_exception(exc_type, exc, traceback)).strip(),
        }



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2016 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Rattail Model Importers
"""

from __future__ import unicode_literals, absolute_import

from sqlalchemy import orm

from rattail.db import model, auth
from rattail.importing import ToSQLAlchemy
from rattail.db.util import format_phone_number, normalize_phone_number


class ToRattail(ToSQLAlchemy):
    """
    Base class for all Rattail model importers.
    """
    key = 'uuid'


class PersonImporter(ToRattail):
    """
    Person data importer.
    """
    model_class = model.Person


class PersonEmailAddressImporter(ToRattail):
    """
    Person email address data importer.
    """
    model_class = model.PersonEmailAddress

    @property
    def supported_fields(self):
        return self.simple_fields + [
            'preferred',
        ]

    def normalize_local_object(self, email):
        data = super(PersonEmailAddressImporter, self).normalize_local_object(email)
        if 'preferred' in self.fields:
            data['preferred'] = email.preference == 1
        return data

    def update_object(self, email, data, local_data=None):
        email = super(PersonEmailAddressImporter, self).update_object(email, data, local_data)
        if 'preferred' in self.fields:
            if data['preferred']:
                if email.preference != 1:
                    person = email.person
                    if not person:
                        person = self.session.query(model.Person).get(email.parent_uuid)
                    if email in person.emails:
                        person.emails.remove(email)
                    person.emails.insert(0, email)
                    person.emails.reorder()
            else:
                if email.preference == 1:
                    person = email.person
                    if not person:
                        person = self.session.query(model.Person).get(email.parent_uuid)
                    if len(person.emails) > 1:
                        person.emails.remove(email)
                        person.emails.append(email)
                        person.emails.reorder()

        # If this is a new record, we may still need to establish its preference.
        if email.preference is None:
            person = email.person
            if not person:
                person = self.session.query(model.Person).get(email.parent_uuid)
            if email not in person.emails:
                person.emails.append(email)
            person.emails.reorder()

        return email


class PersonPhoneNumberImporter(ToRattail):
    """
    Person phone number data importer.
    """
    model_class = model.PersonPhoneNumber

    @property
    def supported_fields(self):
        return self.simple_fields + [
            'normalized_number',
            'preferred',
        ]

    def format_number(self, number):
        return format_phone_number(number)

    def normalize_number(self, number):
        return normalize_phone_number(number)

    def normalize_local_object(self, phone):
        data = super(PersonPhoneNumberImporter, self).normalize_local_object(phone)
        if 'normalized_number' in self.fields:
            data['normalized_number'] = self.normalize_number(phone.number)
        if 'preferred' in self.fields:
            data['preferred'] = phone.preference == 1
        return data

    def update_object(self, phone, data, local_data=None):
        phone = super(PersonPhoneNumberImporter, self).update_object(phone, data, local_data)
        if 'preferred' in self.fields:
            if data['preferred']:
                if phone.preference != 1:
                    person = phone.person
                    if not person:
                        person = self.session.query(model.Person).get(phone.parent_uuid)
                    if phone in person.phones:
                        person.phones.remove(phone)
                    person.phones.insert(0, phone)
                    person.phones.reorder()
            else:
                if phone.preference == 1:
                    person = phone.person
                    if not person:
                        person = self.session.query(model.Person).get(phone.parent_uuid)
                    if len(person.phones) > 1:
                        person.phones.remove(phone)
                        person.phones.append(phone)
                        person.phones.reorder()

        # If this is a new record, we may still need to establish its preference.
        if phone.preference is None:
            person = phone.person
            if not person:
                person = self.session.query(model.Person).get(phone.parent_uuid)
            if phone not in person.phones:
                person.phones.append(phone)
            person.phones.reorder()

        return phone


class PersonMailingAddressImporter(ToRattail):
    """
    Person mailing address data importer.
    """
    model_class = model.PersonMailingAddress


class UserImporter(ToRattail):
    """
    User data importer.
    """
    model_class = model.User


class AdminUserImporter(UserImporter):
    """
    User data importer, plus 'admin' boolean field.
    """

    @property
    def supported_fields(self):
        return super(AdminUserImporter, self).supported_fields + ['admin']

    def get_admin(self, session=None):
        return auth.administrator_role(session or self.session)

    def normalize_local_object(self, user):
        data = super(AdminUserImporter, self).normalize_local_object(user)
        if 'admin' in self.fields:
            data['admin'] = self.get_admin() in user.roles
        return data

    def update_object(self, user, data, local_data=None):
        user = super(UserImporter, self).update_object(user, data, local_data)
        if user:
            if 'admin' in self.fields:
                admin = self.get_admin()
                if data['admin']:
                    if admin not in user.roles:
                        user.roles.append(admin)
                else:
                    if admin in user.roles:
                        user.roles.remove(admin)
            return user


class MessageImporter(ToRattail):
    """
    User message data importer.
    """
    model_class = model.Message


class MessageRecipientImporter(ToRattail):
    """
    User message recipient data importer.
    """
    model_class = model.MessageRecipient


class StoreImporter(ToRattail):
    """
    Store data importer.
    """
    model_class = model.Store


class StorePhoneNumberImporter(ToRattail):
    """
    Store phone data importer.
    """
    model_class = model.StorePhoneNumber


class EmployeeImporter(ToRattail):
    """
    Employee data importer.
    """
    model_class = model.Employee
    person_fields = [
        'first_name',
        'last_name',
        'full_name',
    ]

    @property
    def supported_fields(self):
        return self.simple_fields + self.person_fields + [
            'phone_number',
        ]

    @property
    def person_fields_active(self):
        for field in self.person_fields:
            if field in self.fields:
                return True
        return False

    def cache_query_options(self):
        options = []
        if self.person_fields_active:
            options.append(orm.joinedload(model.Employee.person))
        if 'phone_number' in self.fields:
            options.append(orm.joinedload(model.Employee.phones))
        return options

    def normalize_local_object(self, employee):
        data = super(EmployeeImporter, self).normalize_local_object(employee)

        if self.person_fields_active:
            person = employee.person
            if 'first_name' in self.fields:
                data['first_name'] = person.first_name
            if 'last_name' in self.fields:
                data['last_name'] = person.last_name
            if 'full_name' in self.fields:
                data['full_name'] = person.display_name

        if 'phone_number' in self.fields:
            data['phone_number'] = None
            for phone in employee.phones:
                if phone.type == 'Home':
                    data['phone_number'] = phone.number
                    break

        return data

    def update_object(self, employee, data, local_data=None):
        employee = super(EmployeeImporter, self).update_object(employee, data, local_data)

        if self.person_fields_active:
            person = employee.person
            if 'first_name' in self.fields and person.first_name != data['first_name']:
                person.first_name = data['first_name']
            if 'last_name' in self.fields and person.last_name != data['last_name']:
                person.last_name = data['last_name']
            if 'full_name' in self.fields and person.display_name != data['full_name']:
                person.display_name = data['full_name']

        if 'phone_number' in self.fields:
            number = data['phone_number']
            if number:
                found = False
                for phone in employee.phones:
                    if phone.type == 'Home':
                        if phone.number != number:
                            phone.number = number
                        found = True
                        break
                if not found:
                    employee.add_phone_number(number, type='Home')
            else:
                for phone in list(employee.phones):
                    if phone.type == 'Home':
                        employee.phones.remove(phone)

        return employee


class EmployeeStoreImporter(ToRattail):
    """
    Employee/store data importer.
    """
    model_class = model.EmployeeStore


class EmployeeDepartmentImporter(ToRattail):
    """
    Employee/department data importer.
    """
    model_class = model.EmployeeDepartment


class EmployeeEmailAddressImporter(ToRattail):
    """
    Employee email data importer.
    """
    model_class = model.EmployeeEmailAddress


class EmployeePhoneNumberImporter(ToRattail):
    """
    Employee phone data importer.
    """
    model_class = model.EmployeePhoneNumber


class ScheduledShiftImporter(ToRattail):
    """
    Imports employee scheduled shifts.
    """
    model_class = model.ScheduledShift


class WorkedShiftImporter(ToRattail):
    """
    Imports shifts worked by employees.
    """
    model_class = model.WorkedShift


class CustomerImporter(ToRattail):
    """
    Customer data importer.
    """
    model_class = model.Customer


class CustomerGroupImporter(ToRattail):
    """
    CustomerGroup data importer.
    """
    model_class = model.CustomerGroup


class CustomerGroupAssignmentImporter(ToRattail):
    """
    CustomerGroupAssignment data importer.
    """
    model_class = model.CustomerGroupAssignment


class CustomerPersonImporter(ToRattail):
    """
    CustomerPerson data importer.
    """
    model_class = model.CustomerPerson


class CustomerEmailAddressImporter(ToRattail):
    """
    Customer email address data importer.
    """
    model_class = model.CustomerEmailAddress


class CustomerPhoneNumberImporter(ToRattail):
    """
    Customer phone number data importer.
    """
    model_class = model.CustomerPhoneNumber


class VendorImporter(ToRattail):
    """
    Vendor data importer.
    """
    model_class = model.Vendor


class VendorEmailAddressImporter(ToRattail):
    """
    Vendor email data importer.
    """
    model_class = model.VendorEmailAddress


class VendorPhoneNumberImporter(ToRattail):
    """
    Vendor phone data importer.
    """
    model_class = model.VendorPhoneNumber


class VendorContactImporter(ToRattail):
    """
    Vendor contact data importer.
    """
    model_class = model.VendorContact


class DepartmentImporter(ToRattail):
    """
    Department data importer.
    """
    model_class = model.Department


class SubdepartmentImporter(ToRattail):
    """
    Subdepartment data importer.
    """
    model_class = model.Subdepartment


class CategoryImporter(ToRattail):
    """
    Category data importer.
    """
    model_class = model.Category


class FamilyImporter(ToRattail):
    """
    Family data importer.
    """
    model_class = model.Family


class ReportCodeImporter(ToRattail):
    """
    ReportCode data importer.
    """
    model_class = model.ReportCode


class DepositLinkImporter(ToRattail):
    """
    Deposit link data importer.
    """
    model_class = model.DepositLink


class TaxImporter(ToRattail):
    """
    Tax data importer.
    """
    model_class = model.Tax


class BrandImporter(ToRattail):
    """
    Brand data importer.
    """
    model_class = model.Brand


class ProductImporter(ToRattail):
    """
    Data importer for :class:`rattail.db.model.Product`.
    """
    model_class = model.Product


class ProductCodeImporter(ToRattail):
    """
    Data importer for :class:`rattail.db.model.ProductCode`.
    """
    model_class = model.ProductCode


class ProductCostImporter(ToRattail):
    """
    Data importer for :class:`rattail.db.model.ProductCost`.
    """
    model_class = model.ProductCost


class ProductPriceImporter(ToRattail):
    """
    Data importer for :class:`rattail.db.model.ProductPrice`.
    """
    model_class = model.ProductPrice



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

from unittest import TestCase

from mock import Mock, patch, call

from rattail.db import model
from rattail.db.util import QuerySequence
from rattail.importing import importers
from rattail.tests import NullProgress, RattailTestCase
from rattail.tests.importing import ImporterTester


class ImporterBattery(ImporterTester):
    """
    Battery of tests which can hopefully be ran for any non-bulk importer.
    """

    def test_import_data_empty(self):
        importer = self.make_importer()
        result = importer.import_data()
        self.assertEqual(result, {})

    def test_import_data_dry_run(self):
        importer = self.make_importer()
        self.assertFalse(importer.dry_run)
        importer.import_data(dry_run=True)
        self.assertTrue(importer.dry_run)

    def test_import_data_create(self):
        importer = self.make_importer()
        with patch.object(importer, 'get_key', lambda k: k):
            with patch.object(importer, 'create_object') as create:
                importer.import_data(host_data=[1, 2, 3])
                self.assertEqual(create.call_args_list, [
                    call(1, 1), call(2, 2), call(3, 3)])

    def test_import_data_max_create(self):
        importer = self.make_importer()
        with patch.object(importer, 'get_key', lambda k: k):
            with patch.object(importer, 'create_object') as create:
                importer.import_data(host_data=[1, 2, 3], max_create=1)
                self.assertEqual(create.call_args_list, [call(1, 1)])


class BulkImporterBattery(ImporterBattery):
    """
    Battery of tests which can hopefully be ran for any bulk importer.
    """

    def test_import_data_empty(self):
        importer = self.make_importer()
        result = importer.import_data()
        self.assertEqual(result, 0)


class TestImporter(TestCase):

    def test_init(self):

        # defaults
        importer = importers.Importer()
        self.assertIsNone(importer.model_class)
        self.assertIsNone(importer.key)
        self.assertEqual(importer.fields, [])
        self.assertIsNone(importer.host_system_title)

        # key must be included among the fields
        self.assertRaises(ValueError, importers.Importer, key='upc', fields=[])
        importer = importers.Importer(key='upc', fields=['upc'])
        self.assertEqual(importer.key, ('upc',))
        self.assertEqual(importer.fields, ['upc'])

        # extra bits are passed as-is
        importer = importers.Importer()
        self.assertFalse(hasattr(importer, 'extra_bit'))
        extra_bit = object()
        importer = importers.Importer(extra_bit=extra_bit)
        self.assertIs(importer.extra_bit, extra_bit)

    def test_delete_flag(self):
        # disabled by default
        importer = importers.Importer()
        self.assertTrue(importer.allow_delete)
        self.assertFalse(importer.delete)
        importer.import_data(host_data=[])
        self.assertFalse(importer.delete)

        # but can be enabled
        importer = importers.Importer(delete=True)
        self.assertTrue(importer.allow_delete)
        self.assertTrue(importer.delete)
        importer = importers.Importer()
        self.assertFalse(importer.delete)
        importer.import_data(host_data=[], delete=True)
        self.assertTrue(importer.delete)

    def test_get_host_objects(self):
        importer = importers.Importer()
        objects = importer.get_host_objects()
        self.assertEqual(objects, [])

    def test_cache_local_data(self):
        importer = importers.Importer()
        self.assertRaises(NotImplementedError, importer.cache_local_data)

    def test_get_local_object(self):
        importer = importers.Importer()
        self.assertFalse(importer.caches_local_data)
        self.assertRaises(NotImplementedError, importer.get_local_object, None)

        someobj = object()
        with patch.object(importer, 'get_single_local_object', Mock(return_value=someobj)):
            obj = importer.get_local_object('somekey')
        self.assertIs(obj, someobj)

        importer.caches_local_data = True
        importer.cached_local_data = {'somekey': {'object': someobj, 'data': {}}}
        obj = importer.get_local_object('somekey')
        self.assertIs(obj, someobj)

    def test_get_single_local_object(self):
        importer = importers.Importer()
        self.assertRaises(NotImplementedError, importer.get_single_local_object, None)

    def test_get_cache_key(self):
        importer = importers.Importer(key='upc', fields=['upc'])
        obj = {'upc': '00074305001321'}
        normal = {'data': obj}
        key = importer.get_cache_key(obj, normal)
        self.assertEqual(key, ('00074305001321',))

    def test_normalize_cache_object(self):
        importer = importers.Importer()
        obj = {'upc': '00074305001321'}
        with patch.object(importer, 'normalize_local_object', new=lambda obj: obj):
            cached = importer.normalize_cache_object(obj)
        self.assertEqual(cached, {'object': obj, 'data': obj})

    def test_normalize_local_object(self):
        importer = importers.Importer(key='upc', fields=['upc', 'description'])
        importer.simple_fields = importer.fields
        obj = Mock(upc='00074305001321', description="Apple Cider Vinegar")
        data = importer.normalize_local_object(obj)
        self.assertEqual(data, {'upc': '00074305001321', 'description': "Apple Cider Vinegar"})

    def test_update_object(self):
        importer = importers.Importer(key='upc', fields=['upc', 'description'])
        importer.simple_fields = importer.fields
        obj = Mock(upc='00074305001321', description="Apple Cider Vinegar")

        newobj = importer.update_object(obj, {'upc': '00074305001321', 'description': "Apple Cider Vinegar"})
        self.assertIs(newobj, obj)
        self.assertEqual(obj.description, "Apple Cider Vinegar")

        newobj = importer.update_object(obj, {'upc': '00074305001321', 'description': "Apple Cider Vinegar 32oz"})
        self.assertIs(newobj, obj)
        self.assertEqual(obj.description, "Apple Cider Vinegar 32oz")

    def test_normalize_host_data(self):
        importer = importers.Importer(key='upc', fields=['upc', 'description'],
                                          progress=NullProgress)

        data = [
            {'upc': '00074305001161', 'description': "Apple Cider Vinegar 16oz"},
            {'upc': '00074305001321', 'description': "Apple Cider Vinegar 32oz"},
        ]

        host_data = importer.normalize_host_data(host_objects=[])
        self.assertEqual(host_data, [])

        host_data = importer.normalize_host_data(host_objects=data)
        self.assertEqual(host_data, data)

        with patch.object(importer, 'get_host_objects', new=Mock(return_value=data)):
            host_data = importer.normalize_host_data()
        self.assertEqual(host_data, data)

    def test_get_deletion_keys(self):
        importer = importers.Importer()
        self.assertFalse(importer.caches_local_data)
        keys = importer.get_deletion_keys()
        self.assertEqual(keys, set())

        importer.caches_local_data = True
        self.assertIsNone(importer.cached_local_data)
        keys = importer.get_deletion_keys()
        self.assertEqual(keys, set())

        importer.cached_local_data = {'delete-me': object()}
        keys = importer.get_deletion_keys()
        self.assertEqual(keys, set(['delete-me']))

    def test_model_name_attr(self):
        # default is None
        importer = importers.Importer()
        self.assertIsNone(importer.model_name)

        # but may be overridden via init kwarg
        importer = importers.Importer(model_name='Foo')
        self.assertEqual(importer.model_name, 'Foo')

        # or may inherit its value from 'model_class'
        class Foo:
            pass
        importer = importers.Importer(model_class=Foo)
        self.assertEqual(importer.model_name, 'Foo')

    def test_batch_size_attr(self):
        # default is 200
        importer = importers.Importer()
        self.assertEqual(importer.batch_size, 200)

        # but may be overridden via init kwarg
        importer = importers.Importer(batch_size=0)
        self.assertEqual(importer.batch_size, 0)
        importer = importers.Importer(batch_size=42)
        self.assertEqual(importer.batch_size, 42)

        # batch size determines how often flush occurs
        data = [{'id': i} for i in range(1, 101)]
        importer = importers.Importer(model_name='Foo', key='id', fields=['id'], empty_local_data=True)
        with patch.object(importer, 'create_object'): # just mock that out
            with patch.object(importer, 'flush_create_update') as flush:

                # 4 batches @ 33/per
                importer.import_data(host_data=data, batch_size=33)
                self.assertEqual(flush.call_count, 4)
                flush.reset_mock()

                # 3 batches @ 34/per
                importer.import_data(host_data=data, batch_size=34)
                self.assertEqual(flush.call_count, 3)
                flush.reset_mock()

                # 2 batches @ 50/per
                importer.import_data(host_data=data, batch_size=100)
                self.assertEqual(flush.call_count, 2)
                flush.reset_mock()

                # one extra/final flush happens, whenever the total number of
                # changes happens to match the batch size...

                # 1 batch @ 100/per, plus final flush
                importer.import_data(host_data=data, batch_size=100)
                self.assertEqual(flush.call_count, 2)
                flush.reset_mock()

                # 1 batch @ 200/per
                importer.import_data(host_data=data, batch_size=200)
                self.assertEqual(flush.call_count, 1)
                flush.reset_mock()

                # one extra/final flush also happens when batching is disabled

                # 100 "batches" @ 0/per, plus final flush
                importer.import_data(host_data=data, batch_size=0)
                self.assertEqual(flush.call_count, 101)
                flush.reset_mock()


class TestFromQuery(RattailTestCase):

    def test_query(self):
        importer = importers.FromQuery()
        self.assertRaises(NotImplementedError, importer.query)

    def test_get_host_objects(self):
        query = self.session.query(model.Product)
        importer = importers.FromQuery()
        with patch.object(importer, 'query', Mock(return_value=query)):
            objects = importer.get_host_objects()
        self.assertIsInstance(objects, QuerySequence)


class TestBulkImporter(TestCase, BulkImporterBattery):
    importer_class = importers.BulkImporter
        
    def test_batch_size_attr(self):
        # default is 200
        importer = importers.BulkImporter()
        self.assertEqual(importer.batch_size, 200)

        # but may be overridden via init kwarg
        importer = importers.BulkImporter(batch_size=0)
        self.assertEqual(importer.batch_size, 0)
        importer = importers.BulkImporter(batch_size=42)
        self.assertEqual(importer.batch_size, 42)

        # batch size determines how often flush occurs
        data = [{'id': i} for i in range(1, 101)]
        importer = importers.BulkImporter(model_name='Foo', key='id', fields=['id'], empty_local_data=True)
        with patch.object(importer, 'create_object'): # just mock that out
            with patch.object(importer, 'flush_create_update') as flush:

                # 4 batches @ 33/per
                importer.import_data(host_data=data, batch_size=33)
                self.assertEqual(flush.call_count, 4)
                flush.reset_mock()

                # 3 batches @ 34/per
                importer.import_data(host_data=data, batch_size=34)
                self.assertEqual(flush.call_count, 3)
                flush.reset_mock()

                # 2 batches @ 50/per
                importer.import_data(host_data=data, batch_size=100)
                self.assertEqual(flush.call_count, 2)
                flush.reset_mock()

                # one extra/final flush happens, whenever the total number of
                # changes happens to match the batch size...

                # 1 batch @ 100/per, plus final flush
                importer.import_data(host_data=data, batch_size=100)
                self.assertEqual(flush.call_count, 2)
                flush.reset_mock()

                # 1 batch @ 200/per
                importer.import_data(host_data=data, batch_size=200)
                self.assertEqual(flush.call_count, 1)
                flush.reset_mock()

                # one extra/final flush also happens when batching is disabled

                # 100 "batches" @ 0/per, plus final flush
                importer.import_data(host_data=data, batch_size=0)
                self.assertEqual(flush.call_count, 101)
                flush.reset_mock()


######################################################################
# fake importer class, tested mostly for basic coverage
######################################################################

class Product(object):
    upc = None
    description = None


class MockImporter(importers.Importer):
    model_class = Product
    key = 'upc'
    simple_fields = ['upc', 'description']
    supported_fields = simple_fields
    caches_local_data = True
    flush_every_x = 1
    session = Mock()

    def normalize_local_object(self, obj):
        return obj

    def update_object(self, obj, host_data, local_data=None):
        return host_data


class TestMockImporter(ImporterTester, TestCase):
    importer_class = MockImporter

    sample_data = {
        '16oz': {'upc': '00074305001161', 'description': "Apple Cider Vinegar 16oz"},
        '32oz': {'upc': '00074305001321', 'description': "Apple Cider Vinegar 32oz"},
        '1gal': {'upc': '00074305011283', 'description': "Apple Cider Vinegar 1gal"},
    }

    def setUp(self):
        self.importer = self.make_importer()

    def test_create(self):
        local = self.copy_data()
        del local['32oz']
        self.import_data(local_data=local)
        self.assert_import_created('32oz')
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_create_empty(self):
        self.import_data(host_data={}, local_data={})
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_update(self):
        local = self.copy_data()
        local['16oz']['description'] = "wrong description"
        self.import_data(local_data=local)
        self.assert_import_created()
        self.assert_import_updated('16oz')
        self.assert_import_deleted()

    def test_delete(self):
        local = self.copy_data()
        local['bogus'] = {'upc': '00000000000000', 'description': "Delete Me"}
        self.import_data(local_data=local, delete=True)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted('bogus')

    def test_duplicate(self):
        host = self.copy_data()
        host['32oz-dupe'] = host['32oz']
        self.import_data(host_data=host)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_max_create(self):
        local = self.copy_data()
        del local['16oz']
        del local['1gal']
        self.import_data(local_data=local, max_create=1)
        self.assert_import_created('16oz')
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_max_total_create(self):
        local = self.copy_data()
        del local['16oz']
        del local['1gal']
        self.import_data(local_data=local, max_total=1)
        self.assert_import_created('16oz')
        self.assert_import_updated()
        self.assert_import_deleted()

    def test_max_update(self):
        local = self.copy_data()
        local['16oz']['description'] = "wrong"
        local['1gal']['description'] = "wrong"
        self.import_data(local_data=local, max_update=1)
        self.assert_import_created()
        self.assert_import_updated('16oz')
        self.assert_import_deleted()

    def test_max_total_update(self):
        local = self.copy_data()
        local['16oz']['description'] = "wrong"
        local['1gal']['description'] = "wrong"
        self.import_data(local_data=local, max_total=1)
        self.assert_import_created()
        self.assert_import_updated('16oz')
        self.assert_import_deleted()

    def test_max_delete(self):
        local = self.copy_data()
        local['bogus1'] = {'upc': '00000000000001', 'description': "Delete Me"}
        local['bogus2'] = {'upc': '00000000000002', 'description': "Delete Me"}
        self.import_data(local_data=local, delete=True, max_delete=1)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted('bogus1')

    def test_max_total_delete(self):
        local = self.copy_data()
        local['bogus1'] = {'upc': '00000000000001', 'description': "Delete Me"}
        local['bogus2'] = {'upc': '00000000000002', 'description': "Delete Me"}
        self.import_data(local_data=local, delete=True, max_total=1)
        self.assert_import_created()
        self.assert_import_updated()
        self.assert_import_deleted('bogus1')

    def test_max_total_delete_with_changes(self):
        local = self.copy_data()
        del local['16oz']
        local['32oz']['description'] = "wrong"
        local['1gal']['description'] = "wrong"
        local['bogus1'] = {'upc': '00000000000001', 'description': "Delete Me"}
        local['bogus2'] = {'upc': '00000000000002', 'description': "Delete Me"}
        self.import_data(local_data=local, delete=True, max_total=3)
        self.assert_import_created('16oz')
        self.assert_import_updated('32oz', '1gal')
        self.assert_import_deleted()



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import decimal

from rattail.db import model
from rattail.tests import DataTestCase


class TestDepartment(DataTestCase):

    def test_unicode(self):
        dept = model.Department()
        self.assertEqual(unicode(dept), "")

        dept = model.Department(name=b"Grocery")
        self.assertEqual(unicode(dept), "Grocery")


class TestSubdepartment(DataTestCase):

    def test_unicode(self):
        subdept = model.Subdepartment()
        self.assertEqual(unicode(subdept), "")

        subdept = model.Subdepartment(name=b"Canned Goods")
        self.assertEqual(unicode(subdept), "Canned Goods")


class TestCategory(DataTestCase):

    def test_unicode(self):
        category = model.Category()
        self.assertEqual(unicode(category), "")

        category = model.Category(name=b"Various Odds and Ends")
        self.assertEqual(unicode(category), "Various Odds and Ends")


class TestFamily(DataTestCase):

    def test_unicode(self):
        family = model.Family()
        self.assertEqual(unicode(family), "")

        family = model.Family(name=b"Various Odds and Ends")
        self.assertEqual(unicode(family), "Various Odds and Ends")


class TestReportCode(DataTestCase):

    def test_unicode(self):
        code = model.ReportCode()
        self.assertEqual(unicode(code), "")

        code = model.ReportCode(code=42, name=b"Various Odds and Ends")
        self.assertEqual(unicode(code), "42 - Various Odds and Ends")


class TestDepositLink(DataTestCase):

    def test_unicode(self):
        link = model.DepositLink()
        self.assertEqual(unicode(link), "")

        link = model.DepositLink(description=b"Bottles", amount=decimal.Decimal('0.42'))
        self.assertEqual(unicode(link), "Bottles ($0.42)")



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import csv
import datetime
import argparse
from unittest import TestCase
from cStringIO import StringIO

# from sqlalchemy import func
from mock import patch, Mock
from fixture import TempIO

from rattail.commands import core
from rattail.db import Session, model
from rattail.db.auth import authenticate_user
from rattail.tests import DataTestCase


class TestArgumentParser(TestCase):

    def test_parse_args_preserves_extra_argv(self):
        parser = core.ArgumentParser()
        parser.add_argument('--some-optional-arg')
        parser.add_argument('some_required_arg')
        args = parser.parse_args([
                '--some-optional-arg', 'optional-value', 'required-value',
                'some', 'extra', 'args'])
        self.assertEqual(args.some_required_arg, 'required-value')
        self.assertEqual(args.some_optional_arg, 'optional-value')
        self.assertEqual(args.argv, ['some', 'extra', 'args'])


class TestDateArgument(TestCase):

    def test_valid_date_string_returns_date_object(self):
        date = core.date_argument('2014-01-01')
        self.assertEqual(date, datetime.date(2014, 1, 1))

    def test_invalid_date_string_raises_error(self):
        self.assertRaises(argparse.ArgumentTypeError, core.date_argument, 'invalid-date')


class TestCommand(TestCase):

    def test_initial_subcommands_are_sane(self):
        command = core.Command()
        self.assertTrue('filemon' in command.subcommands)

    def test_unicode(self):
        command = core.Command()
        command.name = 'some-app'
        self.assertEqual(unicode(command), u'some-app')
        
    def test_iter_subcommands_includes_expected_item(self):
        command = core.Command()
        found = False
        for subcommand in command.iter_subcommands():
            if subcommand.name == 'filemon':
                found = True
                break
        self.assertTrue(found)

    def test_print_help(self):
        command = core.Command()
        stdout = StringIO()
        command.stdout = stdout
        command.print_help()
        output = stdout.getvalue()
        stdout.close()
        self.assertTrue('Usage:' in output)
        self.assertTrue('Options:' in output)

    def test_run_with_no_args_prints_help(self):
        command = core.Command()
        with patch.object(command, 'print_help') as print_help:
            command.run()
            print_help.assert_called_once_with()

    def test_run_with_single_help_arg_prints_help(self):
        command = core.Command()
        with patch.object(command, 'print_help') as print_help:
            command.run('help')
            print_help.assert_called_once_with()

    def test_run_with_help_and_unknown_subcommand_args_prints_help(self):
        command = core.Command()
        with patch.object(command, 'print_help') as print_help:
            command.run('help', 'invalid-subcommand-name')
            print_help.assert_called_once_with()

    def test_run_with_help_and_subcommand_args_prints_subcommand_help(self):
        command = core.Command()
        fake = command.subcommands['fake'] = Mock()
        command.run('help', 'fake')
        fake.return_value.parser.print_help.assert_called_once_with()

    def test_run_with_unknown_subcommand_arg_prints_help(self):
        command = core.Command()
        with patch.object(command, 'print_help') as print_help:
            command.run('invalid-command-name')
            print_help.assert_called_once_with()

    def test_stdout_may_be_redirected(self):
        class Fake(core.Subcommand):
            def run(self, args):
                self.stdout.write("standard output stuff")
                self.stdout.flush()
        command = core.Command()
        fake = command.subcommands['fake'] = Fake
        tmp = TempIO()
        config_path = tmp.putfile('test.ini', '')
        out_path = tmp.putfile('out.txt', '')
        command.run('fake', '--config', config_path, '--stdout', out_path)
        with open(out_path) as f:
            self.assertEqual(f.read(), "standard output stuff")

    def test_stderr_may_be_redirected(self):
        class Fake(core.Subcommand):
            def run(self, args):
                self.stderr.write("standard error stuff")
                self.stderr.flush()
        command = core.Command()
        fake = command.subcommands['fake'] = Fake
        tmp = TempIO()
        config_path = tmp.putfile('test.ini', '')
        err_path = tmp.putfile('err.txt', '')
        command.run('fake', '--config', config_path, '--stderr', err_path)
        with open(err_path) as f:
            self.assertEqual(f.read(), "standard error stuff")

    # # TODO: Figure out a better way to test this, or don't bother.
    # def test_noinit_flag_means_no_config(self):
    #     command = commands.Command()
    #     fake = command.subcommands['fake'] = Mock()
    #     command.run('fake', '--no-init')
    #     self.assertEqual(len(fake.return_value.config.files_requested), 0)


class TestSubcommand(TestCase):

    def test_repr(self):
        command = core.Command()
        subcommand = core.Subcommand(command)
        subcommand.name = 'fake-command'
        self.assertEqual(repr(subcommand), "Subcommand(name=u'fake-command')")

    def test_add_parser_args_does_nothing(self):
        command = core.Command()
        subcommand = core.Subcommand(command)
        # Not sure this is really the way to test this, but...
        self.assertEqual(len(subcommand.parser._action_groups[0]._actions), 1)
        subcommand.add_parser_args(subcommand.parser)
        self.assertEqual(len(subcommand.parser._action_groups[0]._actions), 1)

    def test_run_not_implemented(self):
        command = core.Command()
        subcommand = core.Subcommand(command)
        args = subcommand.parser.parse_args([])
        self.assertRaises(NotImplementedError, subcommand.run, args)


class TestAddUser(DataTestCase):
    
    def setUp(self):
        super(TestAddUser, self).setUp()
        self.tmp = TempIO()
        self.stdout_path = self.tmp.putfile('stdout.txt', '')
        self.stderr_path = self.tmp.putfile('stderr.txt', '')

    def test_no_user_created_if_username_already_exists(self):
        self.session.add(model.User(username='fred'))
        self.session.commit()
        self.assertEqual(self.session.query(model.User).count(), 1)
        core.main('adduser', '--no-init', '--stderr', self.stderr_path, 'fred')
        with open(self.stderr_path) as f:
            self.assertEqual(f.read(), "User 'fred' already exists.\n")
        self.assertEqual(self.session.query(model.User).count(), 1)

    # TODO: this breaks when postgres used for test db backend?
    # def test_no_user_created_if_password_prompt_is_canceled(self):
    #     self.assertEqual(self.session.query(model.User).count(), 0)
    #     with patch('rattail.commands.core.getpass') as getpass:
    #         getpass.side_effect = KeyboardInterrupt
    #         core.main('adduser', '--no-init', '--stderr', self.stderr_path, 'fred')
    #     with open(self.stderr_path) as f:
    #         self.assertEqual(f.read(), "\nOperation was canceled.\n")
    #     self.assertEqual(self.session.query(model.User).count(), 0)

    def test_normal_user_created_with_correct_password_but_no_admin_role(self):
        self.assertEqual(self.session.query(model.User).count(), 0)
        with patch('rattail.commands.core.getpass') as getpass:
            getpass.return_value = 'fredpass'
            core.main('adduser', '--no-init', '--stdout', self.stdout_path, 'fred')
        with open(self.stdout_path) as f:
            self.assertEqual(f.read(), "Created user: fred\n")
        fred = self.session.query(model.User).one()
        self.assertEqual(fred.username, 'fred')
        self.assertEqual(len(fred.roles), 0)
        user = authenticate_user(self.session, 'fred', 'fredpass')
        self.assertTrue(user is fred)

    def test_admin_user_created_with_administrator_role(self):
        self.assertEqual(self.session.query(model.User).count(), 0)
        with patch('rattail.commands.core.getpass') as getpass:
            getpass.return_value = 'fredpass'
            core.main('adduser', '--no-init', '--stdout', self.stdout_path, 'fred', '--administrator')
        fred = self.session.query(model.User).one()
        self.assertEqual(len(fred.roles), 1)
        self.assertEqual(fred.roles[0].name, 'Administrator')


class TestDump(DataTestCase):

    def setUp(self):
        super(TestDump, self).setUp()
        self.session.add(model.Product(upc='074305001321'))
        self.session.add(model.Product(upc='074305001161'))
        self.session.commit()

    def test_unknown_model_cannot_be_dumped(self):
        tmp = TempIO()
        stderr_path = tmp.putfile('stderr.txt', '')
        self.assertRaises(SystemExit, core.main, '--no-init', '--stderr', stderr_path, 'dump', 'NoSuchModel')
        with open(stderr_path) as f:
            self.assertEqual(f.read(), "Unknown model: NoSuchModel\n")

    def test_dump_goes_to_stdout_by_default(self):
        tmp = TempIO()
        stdout_path = tmp.putfile('stdout.txt', '')
        core.main('--no-init', '--stdout', stdout_path, 'dump', 'Product')
        with open(stdout_path, 'rb') as csv_file:
            reader = csv.DictReader(csv_file)
            upcs = [row['upc'] for row in reader]
        self.assertEqual(len(upcs), 2)
        self.assertTrue('00074305001321' in upcs)
        self.assertTrue('00074305001161' in upcs)

    def test_dump_goes_to_file_if_so_invoked(self):
        tmp = TempIO()
        output_path = tmp.putfile('output.txt', '')
        core.main('--no-init', 'dump', 'Product', '--output', output_path)
        with open(output_path, 'rb') as csv_file:
            reader = csv.DictReader(csv_file)
            upcs = [row['upc'] for row in reader]
        self.assertEqual(len(upcs), 2)
        self.assertTrue('00074305001321' in upcs)
        self.assertTrue('00074305001161' in upcs)


# TODO: more broken tests..ugh.  these aren't very good or else i might bother
# fixing them...
# class TestFileMonitor(TestCase):

#     @patch('rattail.filemon.linux.start_daemon')
#     def test_start_daemon_with_default_args(self, start_daemon):
#         commands.main('filemon', '--no-init', 'start')
#         start_daemon.assert_called_once_with(None, None, True)

#     @patch('rattail.filemon.linux.start_daemon')
#     def test_start_daemon_with_explicit_args(self, start_daemon):
#         tmp = TempIO()
#         pid_path = tmp.putfile('test.pid', '')
#         commands.main('filemon', '--no-init', '--pidfile', pid_path, '--do-not-daemonize', 'start')
#         start_daemon.assert_called_once_with(None, pid_path, False)

#     @patch('rattail.filemon.linux.stop_daemon')
#     def test_stop_daemon_with_default_args(self, stop_daemon):
#         commands.main('filemon', '--no-init', 'stop')
#         stop_daemon.assert_called_once_with(None, None)

#     @patch('rattail.filemon.linux.stop_daemon')
#     def test_stop_daemon_with_explicit_args(self, stop_daemon):
#         tmp = TempIO()
#         pid_path = tmp.putfile('test.pid', '')
#         commands.main('filemon', '--no-init', '--pidfile', pid_path, 'stop')
#         stop_daemon.assert_called_once_with(None, pid_path)

#     @patch('rattail.commands.sys')
#     def test_unknown_platform_not_supported(self, sys):
#         tmp = TempIO()
#         stderr_path = tmp.putfile('stderr.txt', '')
#         sys.platform = 'bogus'
#         commands.main('--no-init', '--stderr', stderr_path, 'filemon', 'start')
#         sys.exit.assert_called_once_with(1)
#         with open(stderr_path) as f:
#             self.assertEqual(f.read(), "File monitor is not supported on platform: bogus\n")


# # TODO: The purge-batches command tests don't work yet; the db.batches.util
# # tests need to be figured out first...
# class TestPurgeBatches(DataTestCase):

#     def setUp(self):
#         super(TestPurgeBatches, self).setUp()
#         self.session.add(model.Batch(purge=datetime.date(2014, 1, 1)))
#         self.session.add(model.Batch(purge=datetime.date(2014, 2, 1)))
#         self.session.add(model.Batch(purge=datetime.date(2014, 3, 1)))
#         self.session.commit()
#         self.tmp = TempIO()
#         self.stdout_path = self.tmp.putfile('stdout.txt', '')

#     def test_purging_honors_batch_purge_dates(self):
#         self.assertEqual(self.session.query(model.Batch).count(), 3)
#         commands.main('--no-init', '--stdout', self.stdout_path, 'purge-batches', '--date', '2014-01-15')
#         self.assertEqual(self.session.query(model.Batch).count(), 2)
#         self.assertEqual(self.session.query(func.min(model.Batch.purge)).scalar(), datetime.date(2014, 2, 1))
#         with open(self.stdout_path) as f:
#             self.assertTrue(f.read().endswith("\nPurged 1 normal and 0 orphaned batches.\n"))

#     def test_specifying_all_purges_everything(self):
#         self.assertEqual(self.session.query(model.Batch).count(), 3)
#         commands.main('--no-init', '--stdout', self.stdout_path, 'purge-batches', '--all')
#         self.assertEqual(self.session.query(model.Batch).count(), 0)
#         with open(self.stdout_path) as f:
#             self.assertTrue(f.read().endswith("\nPurged 3 normal and 0 orphaned batches.\n"))

#     def test_orphaned_tables_are_also_purged(self):
#         self.session.delete(self.session.query(model.Batch).first())
#         self.session.commit()
#         self.assertEqual(self.session.query(model.Batch).count(), 2)
#         commands.main('--no-init', '--stdout', self.stdout_path, 'purge-batches', '--date', '2013-12-31')
#         self.assertEqual(self.session.query(model.Batch).count(), 2)
#         with open(self.stdout_path) as f:
#             self.assertTrue(f.read().endswith("\nPurged 0 normal and 1 orphaned batches.\n"))



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Fabric Library for SSH
"""

from __future__ import unicode_literals
from __future__ import absolute_import

from fabric.api import *
from fabric.contrib.files import exists, sed

from rattail.fablib import mkdir
from rattail.fablib.python import cdvirtualenv


def configure_ssh(restrict_root=True):
    """
    Configure the OpenSSH service.
    """
    if restrict_root:
        sed('/etc/ssh/sshd_config', r'^PermitRootLogin yes$', r'PermitRootLogin no', use_sudo=True)
    sed('/etc/ssh/sshd_config', r'^#PasswordAuthentication yes$', r'PasswordAuthentication no', use_sudo=True)
    sudo('service ssh restart')


def establish_identity(envname, comment, user='rattail', home='/var/lib/rattail'):
    """
    Generate a SSH key pair and configure it for local use.
    """
    home = home.rstrip('/')
    sshdir = '{0}/.ssh'.format(home)
    owner='{0}:{0}'.format(user)
    mkdir(sshdir, owner=owner, mode='0700')
    with cd(sshdir):
        if not exists('authorized_keys'):
            sudo('touch authorized_keys')
        sudo('chown {0} authorized_keys'.format(owner))
        sudo('chmod 0600 authorized_keys')
    with cdvirtualenv(envname, 'app'):
        mkdir('ssh', owner=owner, mode='0700')
    with cdvirtualenv(envname, 'app/ssh'):
        if not exists('id_rsa', use_sudo=True):
            sudo("ssh-keygen -C '{0}' -P '' -f id_rsa".format(comment))
            sudo('cat id_rsa.pub >> {0}/authorized_keys'.format(sshdir))
        sudo('chown {0} id_rsa*'.format(owner))



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2015 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Organization etc.
"""

from __future__ import unicode_literals

import sqlalchemy as sa
from sqlalchemy.orm import relationship

from .core import Base, uuid_column


class Department(Base):
    """
    Represents an organizational department.
    """
    __tablename__ = 'department'
    __versioned__ = {}

    uuid = uuid_column()
    number = sa.Column(sa.Integer())
    name = sa.Column(sa.String(length=30))

    def __unicode__(self):
        return unicode(self.name or '')


class Subdepartment(Base):
    """
    Represents an organizational subdepartment.
    """
    __tablename__ = 'subdepartment'
    __table_args__ = (
        sa.ForeignKeyConstraint(['department_uuid'], ['department.uuid'], name='subdepartment_fk_department'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    number = sa.Column(sa.Integer())
    name = sa.Column(sa.String(length=30))
    department_uuid = sa.Column(sa.String(length=32))

    def __unicode__(self):
        return self.name or ''


Department.subdepartments = relationship(
    Subdepartment,
    backref='department',
    order_by=Subdepartment.name)


class Category(Base):
    """
    Represents an organizational category for products.
    """
    __tablename__ = 'category'
    __table_args__ = (
        sa.ForeignKeyConstraint(['department_uuid'], ['department.uuid'], name='category_fk_department'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    number = sa.Column(sa.Integer())
    name = sa.Column(sa.String(length=50))
    department_uuid = sa.Column(sa.String(length=32))

    department = relationship(Department)

    def __unicode__(self):
        return self.name or ''


class Family(Base):
    """
    Represents an organizational family for products.
    """
    __tablename__ = 'family'
    __versioned__ = {}

    uuid = uuid_column()
    code = sa.Column(sa.Integer())
    name = sa.Column(sa.String(length=50))

    products = relationship(
        u'Product', back_populates=u'family', doc=u"""\
Collection of :class:`Product` instances which associate with this family.
""")

    def __unicode__(self):
        return unicode(self.name or '')


class ReportCode(Base):
    """
    Represents an organizational "report code" for products.
    """
    __tablename__ = 'report_code'
    __table_args__ = (
        sa.UniqueConstraint('code', name='report_code_uq_code'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    code = sa.Column(sa.Integer(), nullable=False)
    name = sa.Column(sa.String(length=100), nullable=True)

    products = relationship(
        u'Product', back_populates=u'report_code', doc=u"""\
Collection of :class:`Product` instances which associate with this report code.
""")

    def __unicode__(self):
        if not self.code:
            return ''
        return "{} - {}".format(self.code, self.name or '')


class DepositLink(Base):
    """
    Represents a bottle or similar deposit, to which an item may be linked.
    """
    __tablename__ = 'deposit_link'
    __table_args__ = (
        sa.UniqueConstraint('code', name='deposit_link_uq_code'),
        )
    __versioned__ = {}

    uuid = uuid_column()
    code = sa.Column(sa.Integer(), nullable=False)
    description = sa.Column(sa.String(length=255), nullable=True)
    amount = sa.Column(sa.Numeric(precision=5, scale=2), nullable=False)

    def __unicode__(self):
        if not (self.description and self.amount):
            return ""
        return "{0} (${1:0.2f})".format(self.description, self.amount)
